#region UICode
int Amount1 = 2; // [0,200] Radius
bool Amount2 = false; // [0,1] Slider controls \u03C3 instead of radius
bool Amount3 = false; // [0,1] Tileable
byte Amount4 = 0; // Channels to blur|RGB + Alpha|RGB only|Alpha only
#endregion

private float lastSigma = -1.0f;
private float[] lastKernel;

// Build a 1D Gaussian kernel and save lastSigma, lastKernel
private float[] getKernel(float sigma)
{
    // Return cached copy of kernel if it was generated with this sigma value
    if(sigma == lastSigma) 
        return lastKernel;

    int k = (int) Math.Ceiling(3.0f * sigma);
    int kernelSize = (k << 1) + 1;
    
    float[] kernel = new float[kernelSize];
    float total = 0.0f;
    
    // Build kernel and accumulate integral
    for (int i = 0; i < kernelSize; i++)
    {
        kernel[i] = (float) Math.Exp((-0.5 * Math.Pow((double) (i - k), 2.0)) * Math.Pow(sigma, -2.0));
        total += kernel[i];
    }
    
    // Normalize kernel so that intensity of image is unchanged by application
    for (int j = 0; j < kernelSize; j++)
    {
        kernel[j] /= total;
    }
    
    // Save key and value in cache
    lastKernel = kernel;
    lastSigma = sigma;
    
    return lastKernel;
}

// Render the effect for a single region of interest
private void Render(Surface dst, Surface src, Rectangle rect)
{   
    // Get the bounds of the canvas
    Rectangle bounds = src.Bounds;
        
    // Calculate sigma based upon user-supplied radius, or use radius as sigma if requested
    float sigma = !this.Amount2 ? (0.4083f * this.Amount1) + 0.3939f : this.Amount1;
    
    // Calculate maximum window offset (k) using sigma
    int k = (int) Math.Ceiling(3.0f * sigma);
    int kernelSize = (k << 1) + 1;
    
    // Kernel is cached so that subsequent calls for same sigma do not require rebuild
    float[] kernel = getKernel(sigma);

    ColorBgra sourcePixel;
    
    bool blurRGB = (Amount4 == 0 || Amount4 == 1);
    bool blurAlpha = (Amount4 == 0 || Amount4 == 2);
    
    int naiveScore = (rect.Right - rect.Left) * kernelSize;
    int improvedScore = ((rect.Right - rect.Left) << 1) + (k << 1);
    
    int xSrc, ySrc, xDst, yDst;
    int max_x = (bounds.Right - 1);
    int max_y = (bounds.Bottom - 1);
    
    // Adaptively choose convolution algorithm based on estimates of required operations for each    
    if(improvedScore < naiveScore)
    {
        // Run two 1-dimensional convolutions
        // Vertical pass is first because it should outperform horizontal first in most situations
        
        // I didn't know what the Surface constructor was, this worked out fine
        float[,,] intermediate = new float[rect.Right - rect.Left + (k << 1),rect.Bottom - rect.Top,4];
        
        // Vertical pass
        for (int i = rect.Left - k; i < rect.Right + k; i++)
        {
            xSrc = i;
            
            if(!Amount3) // Reflect at edges
            {
                xSrc = (int) Math.Abs(xSrc);
                // TODO: Calculates position as if only one reflection occurs
                if(xSrc > max_x) xSrc = max_x - (xSrc % bounds.Right);
            }
            else // Tile at edges
            {
                if(xSrc < 0) xSrc = max_x + (xSrc % bounds.Right);                        
                else if(xSrc > max_x) xSrc = (xSrc % bounds.Right);
            }                
            
            // Determine where to store this in intermediate array
            xDst = i + k - rect.Left;
            
            for (int j = rect.Top; j < rect.Bottom; j++)
            {
                float[] response = {0.0f,0.0f,0.0f,0.0f};
                
                for (int n = 0; n < kernelSize; n++)
                {    
                    ySrc = j + n - k;    
                
                    if(!Amount3) // Reflect at edges
                    {
                        ySrc = (int) Math.Abs(ySrc);
                        // TODO: Calculates position as if only one reflection occurs
                        if(ySrc > max_y) ySrc = max_y - (ySrc % bounds.Bottom);
                    }
                    else // Tile at edges
                    {
                        if(ySrc < 0) ySrc = max_y + (ySrc % bounds.Bottom);
                        else if(ySrc > max_y) ySrc = (ySrc % bounds.Bottom);
                    }
                                        
                    if(blurRGB)
                    {
                        response[0] += src[xSrc,ySrc].B * kernel[n];
                        response[1] += src[xSrc,ySrc].G * kernel[n];
                        response[2] += src[xSrc,ySrc].R * kernel[n];
                    }
                    
                    if(blurAlpha)
                        response[3] += src[xSrc,ySrc].A * kernel[n];
                }
                
                // Determine where to store this in intermediate array
                yDst = j - rect.Top;

                if(blurRGB)
                {
                    intermediate[xDst,yDst,0] = response[0];
                    intermediate[xDst,yDst,1] = response[1];
                    intermediate[xDst,yDst,2] = response[2];
                }
                
                if(blurAlpha)
                    intermediate[xDst,yDst,3] = response[3];
            }
        }
        
        // Horizontal pass
        for (int i = rect.Left; i < rect.Right; i++)
        {
            for (int j = rect.Top; j < rect.Bottom; j++)
            {
                float[] response = {0.0f,0.0f,0.0f,0.0f};
                
                ySrc = j - rect.Top;
                
                for (int m = 0; m < kernelSize; m++)
                {    
                    xSrc = i + m - rect.Left;
                    
                    if(blurRGB)
                    {
                        response[0] += intermediate[xSrc,ySrc,0] * kernel[m];
                        response[1] += intermediate[xSrc,ySrc,1] * kernel[m];
                        response[2] += intermediate[xSrc,ySrc,2] * kernel[m];
                    }
                    if(blurAlpha)
                        response[3] += intermediate[xSrc,ySrc,3] * kernel[m];
                }
                
                if(!blurRGB)
                {
                    response[0] = src[i,j].B;
                    response[1] = src[i,j].G;
                    response[2] = src[i,j].R;                  
                }
                if(!blurAlpha)
                    response[3] = src[i,j].A;
                
                // Write the result to the destination surface
                dst[i, j] = ColorBgra.FromBgra((byte) Math.Round(response[0]),
                                               (byte) Math.Round(response[1]),
                                               (byte) Math.Round(response[2]),
                                               (byte) Math.Round(response[3]));
            }
        }    
    }
    else 
    {
        // Fall back on naive 2D convolution when it will perform better
        // This usually occurs when a region of interest is small, tall, and skinny
        for (int i = rect.Left; i < rect.Right; i++)
        {
            for (int j = rect.Top; j < rect.Bottom; j++)
            {
                float[] response = {0.0f,0.0f,0.0f,0.0f};
                
                for (int m = 0; m < kernelSize; m++)
                {
                    xSrc = i + m - k;
                    
                    if(!Amount3) // Reflect at edges
                    {
                        xSrc = (int) Math.Abs(xSrc);
                        if(xSrc > max_x) xSrc = max_x - (xSrc % bounds.Right);
                    }
                    else // Tile at edges
                    {
                        if(xSrc < 0) xSrc = max_x + (xSrc % bounds.Right);                      
                        else if(xSrc > max_x) xSrc = (xSrc % bounds.Right);
                    }
                    
                    for (int n = 0; n < kernelSize; n++)
                    {      
                        ySrc = j + n - k;
                        
                        if(!Amount3) // Reflect at edges
                        {
                            ySrc = (int) Math.Abs(ySrc);
                            // TODO: Calculates position as if only one reflection occurs
                            if(ySrc > max_y) ySrc = max_y - (ySrc % bounds.Bottom);
                        }
                        else // Tile at edges
                        {
                            if(ySrc < 0) ySrc = max_y + (ySrc % bounds.Bottom);
                            else if(ySrc > max_y) ySrc = (ySrc % bounds.Bottom);
                        }
                        
                        if(blurRGB)
                        {
                            response[0] += src[xSrc,ySrc].B * kernel[m] * kernel[n];
                            response[1] += src[xSrc,ySrc].G * kernel[m] * kernel[n];
                            response[2] += src[xSrc,ySrc].R * kernel[m] * kernel[n];
                        }
                        
                        if(blurAlpha)
                            response[3] += src[xSrc,ySrc].A * kernel[m] * kernel[n];
                    }
                }
                
                if(!blurRGB)
                {
                    response[0] = src[i,j].B;
                    response[1] = src[i,j].G;
                    response[2] = src[i,j].R;                  
                }
                if(!blurAlpha)
                    response[3] = src[i,j].A;
                
                dst[i, j] = ColorBgra.FromBgra((byte) Math.Round(response[0]),
                                               (byte) Math.Round(response[1]),
                                               (byte) Math.Round(response[2]),
                                               (byte) Math.Round(response[3]));
            }
        }
    }
}